package ru.dbotthepony.mc.otm.menu.data

import com.mojang.blaze3d.platform.InputConstants
import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.longs.Long2ObjectAVLTreeMap
import net.minecraft.client.Minecraft
import net.minecraft.client.gui.screens.Screen
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.ClickAction
import net.minecraft.world.inventory.ClickType
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.menu.MatteryMenu
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import ru.dbotthepony.mc.otm.network.SetCarriedPacket
import ru.dbotthepony.mc.otm.storage.*
import java.math.BigInteger
import java.util.*
import java.util.function.Supplier

class StackAddPacket(val id: Int, val stack_id: Int, val stack: ItemStack) {
	fun write(buffer: FriendlyByteBuf) {
		buffer.writeInt(id)
		buffer.writeInt(stack_id)
		buffer.writeRegistryId(stack.item)
		buffer.writeInt(stack.count)
		buffer.writeBoolean(stack.shareTag != null)

		if (stack.tag != null) buffer.writeNbt(stack.shareTag)
	}

	fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val get = Minecraft.getInstance().player?.containerMenu ?: return@enqueueWork

			if (get.containerId != id)
				return@enqueueWork

			val view = (get as? INetworkedItemViewSupplier)?.networkedItemView ?: throw IllegalStateException("No such item tracker with id $id")

			if (view.localState.containsKey(stack_id)) {
				throw IllegalStateException("Item tracker $id already has stack with id of $stack_id")
			}

			val state = NetworkedItemView.NetworkedItem(stack_id, stack)
			view.localState[stack_id] = state

			/*val iterator = view.sortedView.iterator().withIndex()
			var lastCompare = 0
			var hit = false

			while (iterator.hasNext()) {
				val (i, existing) = iterator.next()
				val cmp = view.sorter.compare(existing.stack, stack)

				if (cmp != lastCompare && lastCompare != 0) {
					hit = true
					view.sortedView.add(i, state)
				} else if (cmp != lastCompare) {
					lastCompare = cmp
				}
			}

			if (!hit) {*/
				view.sortedView.add(state)
			//}

			view.resort()
		}
	}

	companion object {
		fun read(buffer: FriendlyByteBuf): StackAddPacket {
			val id = buffer.readInt()
			val stack = buffer.readInt()
			val item = buffer.readRegistryIdSafe(Item::class.java)
			val count = buffer.readInt()
			val state = ItemStack(item, count)
			if (buffer.readBoolean()) state.readShareTag(buffer.readNbt())
			return StackAddPacket(id, stack, state)
		}
	}
}

class StackChangePacket(val id: Int, val stackID: Int, val newCount: Int) {
	fun write(buffer: FriendlyByteBuf) {
		buffer.writeInt(id)
		buffer.writeInt(stackID)
		buffer.writeInt(newCount)
	}

	fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val get = Minecraft.getInstance().player?.containerMenu ?: return@enqueueWork

			if (get.containerId != id)
				return@enqueueWork

			val view = (get as? INetworkedItemViewSupplier)?.networkedItemView ?: throw IllegalStateException("No such item tracker with id $id")
			val state = view.localState[stackID] ?: throw IllegalStateException("No such stack with id $stackID in $view")

			state.stack.count = newCount
			view.resort()
		}
	}

	companion object {
		fun read(buffer: FriendlyByteBuf): StackChangePacket {
			val id = buffer.readInt()
			val stackID = buffer.readInt()
			val newCount = buffer.readInt()
			return StackChangePacket(id, stackID, newCount)
		}
	}
}

class StackRemovePacket(val id: Int, val stackID: Int) {
	fun write(buffer: FriendlyByteBuf) {
		buffer.writeInt(id)
		buffer.writeInt(stackID)
	}

	fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val get = Minecraft.getInstance().player?.containerMenu ?: return@enqueueWork

			if (get.containerId != id)
				return@enqueueWork

			val view = (get as? INetworkedItemViewSupplier)?.networkedItemView ?: throw IllegalStateException("No such item tracker with id $id")
			val obj = view.localState.remove(stackID) ?: throw IllegalStateException("No such stack with id $stackID in $view")
			view.sortedView.remove(obj)
			view.resort()
		}
	}

	companion object {
		fun read(buffer: FriendlyByteBuf): StackRemovePacket {
			val id = buffer.readInt()
			val stackID = buffer.readInt()
			return StackRemovePacket(id, stackID)
		}
	}
}

/**
 * Creates a virtual, slotless container for Player to interaction with.
 */
open class NetworkedItemView(val ply: Player, val menu: MatteryMenu, val remote: Boolean) : IStorageEventConsumer<ItemStackWrapper> {
	data class NetworkedItem constructor(val id: Int, val stack: ItemStack, val upstreamId: UUID? = null)

	override val storageType: StorageStackType<ItemStackWrapper>
		get() = ITEM_STORAGE

	// this (how client see and interact with)
	val localState = Int2ObjectAVLTreeMap<NetworkedItem>()

	val sortedView = LinkedList<NetworkedItem>()
	var sorter: Comparator<ItemStack> = NAME_SORTER

	companion object {
		val NAME_SORTER = Comparator<ItemStack> { o1, o2 ->
			val cmp = o1.displayName.string.compareTo(o2.displayName.string)

			if (cmp != 0)
				return@Comparator cmp

			return@Comparator o1.item.registryName.toString().compareTo(o2.item.registryName.toString())
		}

		val COUNT_SORTER = Comparator<ItemStack> { o1, o2 ->
			val cmp = o1.count.compareTo(o2.count)

			if (cmp != 0)
				return@Comparator cmp

			return@Comparator o1.item.registryName.toString().compareTo(o2.item.registryName.toString())
		}
	}

	fun resort() {
		sortedView.sortWith { a, b ->
			return@sortWith sorter.compare(a.stack, b.stack)
		}
	}

	// parent (e.g. VirtualComponent)
	protected val upstreamState = HashMap<UUID, NetworkedItem>()
	protected val networkBacklog = ArrayList<Any>()

	operator fun get(id: Int): NetworkedItem? = localState[id]

	protected var provider: IStorageComponent<ItemStackWrapper>? = null

	fun mouseClick(index: Int, mouse_click_type: Int) {
		val list = sortedView

		val action =
			if (mouse_click_type == InputConstants.MOUSE_BUTTON_LEFT) ClickAction.PRIMARY else ClickAction.SECONDARY

		val type =
			if (mouse_click_type == InputConstants.MOUSE_BUTTON_MIDDLE) ClickType.CLONE else if (Screen.hasShiftDown()) ClickType.QUICK_MOVE else ClickType.PICKUP

		MatteryNetworking.send(
			null,
			InteractPacket(menu.containerId, if (index >= list.size) -1 else list[index].id, type, action)
		)
	}

	fun setComponent(provider: IStorageComponent<ItemStackWrapper>?) {
		if (provider === this.provider) return

		this.provider?.removeListenerAuto(this)
		this.provider = provider
		provider?.addListenerAuto(this)
	}

	fun removed() {
		provider?.removeListenerAuto(this)
	}

	val itemCount get() = localState.values.size

	override fun addStack(stack: ItemStackWrapper, id: UUID, provider: IStorageProvider<ItemStackWrapper>) = addObject(stack.stack, id)
	override fun changeStack(stack: ItemStackWrapper, id: UUID, oldCount: BigInteger) = changeObject(id, stack.count.toInt())

	protected fun network(fn: () -> Any) {
		if (!remote) {
			networkBacklog.add(fn())
		}
	}

	override fun removeStack(stack: ItemStackWrapper, id: UUID) {
		val get = upstreamState[id] ?: throw IllegalStateException("Unknown ItemStack with upstream id $id!")
		upstreamState.remove(id)
		localState.remove(get.id)
		network { StackRemovePacket(menu.containerId, get.id) }
	}

	protected var nextItemID = 0

	fun addObject(stack: ItemStack, id_upstream: UUID) {
		check(!upstreamState.containsKey(id_upstream)) { "Already tracking ItemStack with upstream id $id_upstream!" }

		val state = NetworkedItem(nextItemID++, stack.copy(), id_upstream)

		this.localState[state.id] = state
		upstreamState[id_upstream] = state
		network { StackAddPacket(menu.containerId, state.id, stack) }
	}

	fun changeObject(id_upstream: UUID, new_count: Int) {
		val get = upstreamState[id_upstream] ?: throw IllegalStateException("Unknown ItemStack with upstream id $id_upstream!")
		get.stack.count = new_count
		network { StackChangePacket(menu.containerId, get.id, new_count) }
	}

	fun clear() {
		sortedView.clear()
		upstreamState.clear()
		localState.clear()

		if (!remote) {
			networkBacklog.clear()
			networkBacklog.add(ClearPacket(menu.containerId))
		}
	}

	fun network() {
		check(!remote) { "Not a server" }
		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }
		for (packet in networkBacklog) MatteryNetworking.CHANNEL.send(consumer, packet)
		networkBacklog.clear()
	}

	fun playerInteract(packet: InteractPacket) {
		val provider = provider ?: return

		val click = packet.click
		val action = packet.action
		val stack_id = packet.stack_id

		if (click == ClickType.CLONE) {
			if (stack_id < 0 || !ply.abilities.instabuild) return

			val state = get(stack_id) ?: return
			val copy: ItemStack = state.stack.copy()

			copy.count = Math.min(copy.count, copy.maxStackSize)
			ply.containerMenu.carried = copy
			MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(ply.containerMenu.carried))
			ply.containerMenu.setRemoteCarried(ply.containerMenu.carried.copy())

			return
		}

		if (click == ClickType.QUICK_MOVE && stack_id > -1) {
			val state = get(stack_id) ?: return

			val amount =
				if (action == ClickAction.PRIMARY)
					state.stack.maxStackSize
				else
					Math.max(1, state.stack.maxStackSize / 2)

			val extracted = provider.extractStack(state.upstreamId!!, amount.toBigInteger(), true)

			if (!extracted.isEmpty) {
				val (_, remaining) = menu.quickMoveToInventory(extracted.stack, false)

				if (remaining.count != extracted.count.toInt()) {
					provider.extractStack(state.upstreamId, (extracted.count.toInt() - remaining.count).toBigInteger(), false)
				}
			}

			return
		}

		if (!menu.carried.isEmpty && click != ClickType.QUICK_MOVE) {
			// try to put
			if (action == ClickAction.PRIMARY) {
				val carried = menu.carried
				val amount = carried.count

				if (amount == carried.count) {
					val stack = provider.insertStack(ItemStackWrapper(menu.carried), false).stack
					menu.carried = stack
					MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(menu.carried))
					menu.setRemoteCarried(menu.carried.copy())
				}
			} else {
				val copy = menu.carried.copy()
				copy.count = 1

				if (provider.insertStack(ItemStackWrapper(copy), false).isEmpty) {
					menu.carried.shrink(1)
					MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(menu.carried))
					menu.setRemoteCarried(menu.carried.copy())
				}
			}
		} else if (stack_id > -1) {
			val state = get(stack_id) ?: return

			val amount =
				if (action == ClickAction.PRIMARY)
					state.stack.maxStackSize
				else
					(state.stack.count / 2).coerceAtMost(state.stack.maxStackSize / 2).coerceAtLeast(1)

			val extracted = provider.extractStack(state.upstreamId!!, amount.toBigInteger(), false)
			menu.carried = extracted.stack
			MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(menu.carried))
			menu.setRemoteCarried(menu.carried.copy())
		}
	}
}
